@@ -147,11 +147,11 @@ class ArtifactCache():
     #
     # Args:
     #     element (Element): The Element to check
-    #     strength (_KeyStrength): Either STRONG or WEAK key strength, or None
+    #     key (str): The cache key to use
     #
     # Returns: True if the artifact is in the cache, False otherwise
     #
-    def contains(self, element, strength=None):
+    def contains(self, element, key):
         raise ImplError("Cache '{kind}' does not implement contains()"
                         .format(kind=type(self).__name__))
 
@@ -164,6 +164,7 @@ class ArtifactCache():
     #
     # Args:
     #     element (Element): The Element to extract
+    #     key (str): The cache key to use
     #
     # Raises:
     #     ArtifactError: In cases there was an OSError, or if the artifact
@@ -171,7 +172,7 @@ class ArtifactCache():
     #
     # Returns: path to extracted artifact
     #
-    def extract(self, element):
+    def extract(self, element, key):
         raise ImplError("Cache '{kind}' does not implement extract()"
                         .format(kind=type(self).__name__))
 
@@ -182,8 +183,9 @@ class ArtifactCache():
     # Args:
     #     element (Element): The Element commit an artifact for
     #     content (str): The element's content directory
+    #     keys (list): The cache keys to use
     #
-    def commit(self, element, content):
+    def commit(self, element, content, keys):
         raise ImplError("Cache '{kind}' does not implement commit()"
                         .format(kind=type(self).__name__))
 
@@ -208,16 +210,81 @@ class ArtifactCache():
     def has_push_remotes(self, *, element=None):
         return False
 
-    # remote_contains_key():
+    # remote_contains():
     #
     # Check whether the artifact for the specified Element is already available
     # in any remote artifact cache.
     #
     # Args:
     #     element (Element): The Element to check
-    #     strength (_KeyStrength): Either STRONG or WEAK key strength, or None
+    #     key (str): The cache key to use
     #
-    # Returns: True if the artifact is in the cache, False otherwise
+    # Returns: True if the artifact is in a cache, False otherwise
     #
-    def remote_contains(self, element, strength=None):
+    def remote_contains(self, element, key):
         return False
+
+    # push_needed():
+    #
+    # Check whether an artifact for the specified Element needs to be pushed to
+    # any of the configured push remotes. The policy is to push every artifact
+    # we build to every configured push remote, so this should only return False
+    # if all of the configured push remotes already contain the given artifact.
+    #
+    # This function checks for presence of the artifact only using its strong
+    # key. The presence of the weak key in a cache does not necessarily indicate
+    # that this particular artifact is present, only that there is a
+    # partially-compatible version available.
+    #
+    # Args:
+    #     element (Element): The Element to check
+    #     key (str): The cache key to use
+    #
+    # Returns: False if all the push remotes have the artifact, True otherwise
+    #
+    def push_needed(self, element, key):
+        return False
+
+    # push():
+    #
+    # Push committed artifact to remote repository.
+    #
+    # Args:
+    #     element (Element): The Element whose artifact is to be pushed
+    #     keys (list): The cache keys to use
+    #
+    # Returns:
+    #   (bool): True if any remote was updated, False if no pushes were required
+    #
+    # Raises:
+    #   (ArtifactError): if there was an error
+    #
+    def push(self, element, keys):
+        raise ImplError("Cache '{kind}' does not implement push()"
+                        .format(kind=type(self).__name__))
+
+    # pull():
+    #
+    # Pull artifact from one of the configured remote repositories.
+    #
+    # Args:
+    #     element (Element): The Element whose artifact is to be fetched
+    #     key (str): The cache key to use
+    #     progress (callable): The progress callback, if any
+    #
+    def pull(self, element, key, *, progress=None):
+        raise ImplError("Cache '{kind}' does not implement pull()"
+                        .format(kind=type(self).__name__))
+
+    # link_key():
+    #
+    # Add a key for an existing artifact.
+    #
+    # Args:
+    #     element (Element): The Element whose artifact is to be linked
+    #     oldkey (str): An existing cache key for the artifact
+    #     newkey (str): A new cache key for the artifact
+    #
+    def link_key(self, element, oldkey, newkey):
+        raise ImplError("Cache '{kind}' does not implement link_key()"
+                        .format(kind=type(self).__name__))
