@@ -26,7 +26,6 @@ import tempfile
 
 from .. import _ostree, _signals, utils
 from .._exceptions import ArtifactError
-from ..element import _KeyStrength
 from .._ostree import OSTreeError
 
 from . import ArtifactCache
@@ -52,6 +51,37 @@ def buildref(element, key):
     return '{0}/{1}/{2}'.format(project.name, element_name, key)
 
 
+# Represents a single remote OSTree cache.
+#
+class _OSTreeRemote():
+    def __init__(self, spec, pull_url, push_url):
+        self.spec = spec
+        self.pull_url = pull_url
+        self.push_url = push_url
+
+
+# Maps artifacts to the remotes that contain them.
+#
+class _OSTreeArtifactMap():
+    def __init__(self):
+        self._ref_to_remotes = {}
+
+    def append(self, ref, remote):
+        if ref in self._ref_to_remotes:
+            self._ref_to_remotes[ref].append(remote)
+        else:
+            self._ref_to_remotes[ref] = [remote]
+
+    def lookup(self, ref):
+        return self._ref_to_remotes.get(ref, [])
+
+    def lookup_first(self, ref):
+        return self._ref_to_remotes.get(ref, [])[0]
+
+    def contains(self, ref):
+        return ref in self._ref_to_remotes
+
+
 # An OSTreeCache manages artifacts in an OSTree repository
 #
 # Args:
@@ -65,7 +95,7 @@ def buildref(element, key):
 #
 class OSTreeCache(ArtifactCache):
 
-    def __init__(self, context, enable_push):
+    def __init__(self, context, *, enable_push):
         super().__init__(context)
 
         self.enable_push = enable_push
@@ -73,9 +103,10 @@ class OSTreeCache(ArtifactCache):
         ostreedir = os.path.join(context.artifactdir, 'ostree')
         self.repo = _ostree.ensure(ostreedir, False)
 
-        self.push_urls = {}
-        self.pull_urls = {}
-        self._remote_refs = {}
+        # Per-project list of OSTreeRemote and OSTreeArtifactMap instances.
+        self._remotes = {}
+        self._artifact_maps = {}
+
         self._has_fetch_remotes = False
         self._has_push_remotes = False
 
@@ -91,7 +122,8 @@ class OSTreeCache(ArtifactCache):
             return True
         else:
             # Check whether the specified element's project has push remotes
-            return len(self.push_urls[element._get_project()]) > 0
+            remotes_for_project = self._remotes[element._get_project()]
+            return any(remote.spec.push for remote in remotes_for_project)
 
     # contains():
     #
@@ -100,45 +132,32 @@ class OSTreeCache(ArtifactCache):
     #
     # Args:
     #     element (Element): The Element to check
-    #     strength (_KeyStrength): Either STRONG or WEAK key strength, or None
+    #     key (str): The cache key to use
     #
     # Returns: True if the artifact is in the cache, False otherwise
     #
-    def contains(self, element, strength=None):
-        if strength is None:
-            strength = _KeyStrength.STRONG if element._get_strict() else _KeyStrength.WEAK
-
-        if strength == _KeyStrength.STRONG:
-            key = element._get_strict_cache_key()
-        else:
-            key = element._get_cache_key(strength)
-        if not key:
-            return False
-
+    def contains(self, element, key):
         ref = buildref(element, key)
         return _ostree.exists(self.repo, ref)
 
-    # remote_contains_key():
+    # remotes_containing_key():
     #
-    # Check whether the artifact for the specified Element is already available
-    # in the remote artifact cache.
+    # Return every remote cache that contains the key. The result will be an
+    # ordered list of remotes.
     #
     # Args:
     #     element (Element): The Element to check
     #     key (str): The key to use
     #
-    # Returns: True if the artifact is in the cache, False otherwise
+    # Returns (list): A list of _OSTreeRemote instances.
     #
-    def remote_contains_key(self, element, key):
+    def remotes_containing_key(self, element, key):
         if not self._has_fetch_remotes:
-            return False
-
-        remote_refs = self._remote_refs[element._get_project()]
-        if len(remote_refs) == 0:
-            return False
+            return []
 
+        artifact_map = self._artifact_maps[element._get_project()]
         ref = buildref(element, key)
-        return ref in remote_refs
+        return artifact_map.lookup(ref)
 
     # remote_contains():
     #
@@ -147,22 +166,39 @@ class OSTreeCache(ArtifactCache):
     #
     # Args:
     #     element (Element): The Element to check
-    #     strength (_KeyStrength): Either STRONG or WEAK key strength, or None
+    #     key (str): The cache key to use
     #
-    # Returns: True if the artifact is in the cache, False otherwise
+    # Returns: True if the artifact is in a cache, False otherwise
     #
-    def remote_contains(self, element, strength=None):
-        if strength is None:
-            strength = _KeyStrength.STRONG if element._get_strict() else _KeyStrength.WEAK
+    def remote_contains(self, element, key):
+        remotes = self.remotes_containing_key(element, key)
+        return len(remotes) > 0
 
-        if strength == _KeyStrength.STRONG:
-            key = element._get_strict_cache_key()
-        else:
-            key = element._get_cache_key(strength)
-        if not key:
-            return False
+    # push_needed():
+    #
+    # Check whether an artifact for the specified Element needs to be pushed to
+    # any of the configured push remotes. The policy is to push every artifact
+    # we build to every configured push remote, so this should only return False
+    # if all of the configured push remotes already contain the given artifact.
+    #
+    # This function checks for presence of the artifact only using its strong
+    # key. The presence of the weak key in a cache does not necessarily indicate
+    # that this particular artifact is present, only that there is a
+    # partially-compatible version available.
+    #
+    # Args:
+    #     element (Element): The Element to check
+    #     key (str): The cache key to use
+    #
+    # Returns: False if all the push remotes have the artifact, True otherwise
+    #
+    def push_needed(self, element, key):
 
-        return self.remote_contains_key(element, key)
+        remotes_with_artifact = self.remotes_containing_key(element, key)
+
+        push_remotes_with_artifact = set(r for r in remotes_with_artifact if r.spec.push)
+        push_remotes_for_project = set(self._remotes[element._get_project()])
+        return not (push_remotes_for_project.issubset(push_remotes_with_artifact))
 
     # extract():
     #
@@ -173,6 +209,7 @@ class OSTreeCache(ArtifactCache):
     #
     # Args:
     #     element (Element): The Element to extract
+    #     key (str): The cache key to use
     #
     # Raises:
     #     ArtifactError: In cases there was an OSError, or if the artifact
@@ -180,18 +217,12 @@ class OSTreeCache(ArtifactCache):
     #
     # Returns: path to extracted artifact
     #
-    def extract(self, element):
-        ref = buildref(element, element._get_strict_cache_key())
+    def extract(self, element, key):
+        ref = buildref(element, key)
 
         # resolve ref to checksum
         rev = _ostree.checksum(self.repo, ref)
 
-        # resolve weak cache key, if artifact is missing for strong cache key
-        # and the context allows use of weak cache keys
-        if not rev and not element._get_strict():
-            ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
-            rev = _ostree.checksum(self.repo, ref)
-
         if not rev:
             raise ArtifactError("Artifact missing for {}".format(ref))
 
@@ -229,16 +260,13 @@ class OSTreeCache(ArtifactCache):
     # Args:
     #     element (Element): The Element commit an artifact for
     #     content (str): The element's content directory
+    #     keys (list): The cache keys to use
     #
-    def commit(self, element, content):
-        # tag with strong cache key based on dependency versions used for the build
-        ref = buildref(element, element._get_cache_key())
-
-        # also store under weak cache key
-        weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
+    def commit(self, element, content, keys):
+        refs = [buildref(element, key) for key in keys]
 
         try:
-            _ostree.commit(self.repo, content, ref, weak_ref)
+            _ostree.commit(self.repo, content, refs)
         except OSTreeError as e:
             raise ArtifactError("Failed to commit artifact: {}".format(e)) from e
 
@@ -248,76 +276,75 @@ class OSTreeCache(ArtifactCache):
     #
     # Args:
     #     element (Element): The Element whose artifact is to be fetched
+    #     key (str): The cache key to use
     #     progress (callable): The progress callback, if any
     #
-    def pull(self, element, progress=None):
+    def pull(self, element, key, *, progress=None):
         project = element._get_project()
 
-        remote_refs = self._remote_refs[project]
+        artifact_map = self._artifact_maps[project]
 
-        ref = buildref(element, element._get_strict_cache_key())
-        weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
+        ref = buildref(element, key)
 
         try:
-            if ref in remote_refs:
-                # fetch the artifact using the strong cache key
-                _ostree.fetch(self.repo, remote=remote_refs[ref],
-                              ref=ref, progress=progress)
-
-                # resolve ref to checksum
-                rev = _ostree.checksum(self.repo, ref)
-
-                # update weak ref by pointing it to this newly fetched artifact
-                _ostree.set_ref(self.repo, weak_ref, rev)
-            elif weak_ref in remote_refs:
-                # fetch the artifact using the weak cache key
-                _ostree.fetch(self.repo, remote=remote_refs[weak_ref],
-                              ref=weak_ref, progress=progress)
-
-                # resolve weak_ref to checksum
-                rev = _ostree.checksum(self.repo, weak_ref)
-
-                # extract strong cache key from this newly fetched artifact
-                element._update_state()
-                ref = buildref(element, element._get_cache_key())
-
-                # create tag for strong cache key
-                _ostree.set_ref(self.repo, ref, rev)
-            else:
-                raise ArtifactError("Attempt to pull unavailable artifact for element {}"
-                                    .format(element.name))
+            # fetch the artifact from highest priority remote using the specified cache key
+            remote = artifact_map.lookup_first(ref)
+            remote_name = self._ensure_remote(self.repo, remote.pull_url)
+            _ostree.fetch(self.repo, remote=remote_name, ref=ref, progress=progress)
         except OSTreeError as e:
             raise ArtifactError("Failed to pull artifact for element {}: {}"
                                 .format(element.name, e)) from e
 
+    # link_key():
+    #
+    # Add a key for an existing artifact.
+    #
+    # Args:
+    #     element (Element): The Element whose artifact is to be linked
+    #     oldkey (str): An existing cache key for the artifact
+    #     newkey (str): A new cache key for the artifact
+    #
+    def link_key(self, element, oldkey, newkey):
+        oldref = buildref(element, oldkey)
+        newref = buildref(element, newkey)
+
+        # resolve ref to checksum
+        rev = _ostree.checksum(self.repo, oldref)
+
+        # create additional ref for the same checksum
+        _ostree.set_ref(self.repo, newref, rev)
+
     # push():
     #
     # Push committed artifact to remote repository.
     #
     # Args:
     #     element (Element): The Element whose artifact is to be pushed
+    #     keys (list): The cache keys to use
     #
     # Returns:
-    #   (bool): True if the remote was updated, False if it already existed
-    #           and no updated was required
+    #   (bool): True if any remote was updated, False if no pushes were required
     #
     # Raises:
     #   (ArtifactError): if there was an error
-    def push(self, element):
+    def push(self, element, keys):
         any_pushed = False
 
         project = element._get_project()
 
-        push_urls = self.push_urls[project]
+        push_remotes = [r for r in self._remotes[project] if r.spec.push]
 
-        if len(push_urls) == 0:
+        if len(push_remotes) == 0:
             raise ArtifactError("Push is not enabled for any of the configured remote artifact caches.")
 
-        ref = buildref(element, element._get_cache_key())
-        weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
+        refs = [buildref(element, key) for key in keys]
 
-        for push_url in push_urls:
-            any_pushed |= self._push_to_remote(push_url, element, ref, weak_ref)
+        remotes_for_each_ref = [self.remotes_containing_key(element, ref) for ref in refs]
+
+        for remote in push_remotes:
+            # Push if the remote is missing any of the refs
+            if any([remote not in remotes_with_ref for remotes_with_ref in remotes_for_each_ref]):
+                any_pushed |= self._push_to_remote(remote, element, refs)
 
         return any_pushed
 
@@ -409,8 +436,8 @@ class OSTreeCache(ArtifactCache):
         # possible to pickle local functions such as child_action().
         #
         q = multiprocessing.Queue()
-        for remote in remote_specs:
-            p = multiprocessing.Process(target=child_action, args=(remote.url, q))
+        for remote_spec in remote_specs:
+            p = multiprocessing.Process(target=child_action, args=(remote_spec.url, q))
 
             try:
 
@@ -425,16 +452,16 @@ class OSTreeCache(ArtifactCache):
                 raise
 
             if error and on_failure:
-                on_failure(remote.url, error)
+                on_failure(remote_spec.url, error)
             elif error:
                 raise ArtifactError(error)
             else:
-                if remote.push and push_url:
+                if remote_spec.push and push_url:
                     self._has_push_remotes = True
                 if pull_url:
                     self._has_fetch_remotes = True
 
-                remote_results[remote.url] = (push_url, pull_url, remote_refs)
+                remote_results[remote_spec.url] = (push_url, pull_url, remote_refs)
 
         # Prepare push_urls, pull_urls, and remote_refs for each project
         for project in self.context._get_projects():
@@ -442,43 +469,33 @@ class OSTreeCache(ArtifactCache):
             if project in self.project_remote_specs:
                 remote_specs = list(utils._deduplicate(remote_specs + self.project_remote_specs[project]))
 
-            push_urls = []
-            pull_urls = []
-            _remote_refs = {}
+            remotes = []
+            artifact_map = _OSTreeArtifactMap()
 
-            for remote in remote_specs:
+            for remote_spec in remote_specs:
                 # Errors are already handled in the loop above,
                 # skip unreachable remotes here.
-                if remote.url not in remote_results:
+                if remote_spec.url not in remote_results:
                     continue
 
-                push_url, pull_url, remote_refs = remote_results[remote.url]
+                push_url, pull_url, remote_refs = remote_results[remote_spec.url]
 
-                if remote.push:
-                    if push_url:
-                        push_urls.append(push_url)
-                    else:
-                        raise ArtifactError("Push enabled but not supported by repo at: {}".format(remote.url))
+                if remote_spec.push and not push_url:
+                    raise ArtifactError("Push enabled but not supported by repo at: {}".format(remote_spec.url))
 
-                # The specs are deduplicated when reading the config, but since
-                # each push URL can supply an arbitrary pull URL we must dedup
-                # those again here.
-                if pull_url and pull_url not in pull_urls:
-                    pull_urls.append(pull_url)
+                remote = _OSTreeRemote(remote_spec, pull_url, push_url)
+                remotes.append(remote)
 
                 # Update our overall map of remote refs with any refs that are
                 # present in the new remote and were not already found in
                 # higher priority ones.
-                remote = self._ensure_remote(self.repo, pull_url)
                 for ref in remote_refs:
-                    if ref not in _remote_refs:
-                        _remote_refs[ref] = remote
+                    artifact_map.append(ref, remote)
 
-            self.push_urls[project] = push_urls
-            self.pull_urls[project] = pull_urls
-            self._remote_refs[project] = _remote_refs
+            self._artifact_maps[project] = artifact_map
+            self._remotes[project] = remotes
 
-    def _push_to_remote(self, push_url, element, ref, weak_ref):
+    def _push_to_remote(self, remote, element, refs):
         with utils._tempdir(dir=self.context.artifactdir, prefix='push-repo-') as temp_repo_dir:
 
             with element.timed_activity("Preparing compressed archive"):
@@ -487,16 +504,16 @@ class OSTreeCache(ArtifactCache):
                 temp_repo = _ostree.ensure(temp_repo_dir, True)
 
                 # Now push the ref we want to push into our temporary archive-z2 repo
-                _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=ref)
-                _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=weak_ref)
+                for ref in refs:
+                    _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=ref)
 
             with element.timed_activity("Sending artifact"), \
                 element._output_file() as output_file:
                 try:
                     pushed = push_artifact(temp_repo.get_path().get_path(),
-                                           push_url,
-                                           [ref, weak_ref], output_file)
+                                           remote.push_url,
+                                           refs, output_file)
                 except PushException as e:
-                    raise ArtifactError("Failed to push artifact {}: {}".format(ref, e)) from e
+                    raise ArtifactError("Failed to push artifact {}: {}".format(refs, e)) from e
 
             return pushed
