@@ -147,7 +147,7 @@ click.BaseCommand.main = override_main
               help="Force enable/disable ANSI color codes in output")
 @click.option('--strict/--no-strict', default=None, is_flag=True,
               help="Elements must be rebuilt when their dependencies have changed")
-@click.option('--option', '-o', type=click.Tuple([str, str]), multiple=True,
+@click.option('--option', '-o', type=click.Tuple([str, str]), multiple=True, metavar='OPTION VALUE',
               help="Specify a project option")
 @click.pass_context
 def cli(context, **kwargs):
@@ -458,20 +458,25 @@ def show(app, elements, deps, except_, order, format, downloadable):
 ##################################################################
 @cli.command(short_help="Shell into an element's sandbox environment")
 @click.option('--build', '-b', is_flag=True, default=False,
-              help='Create a build sandbox')
+              help='Stage dependencies and sources to build')
 @click.option('--sysroot', '-s', default=None,
               type=click.Path(exists=True, file_okay=False, readable=True),
               help="An existing sysroot")
+@click.option('--mount', type=click.Tuple([click.Path(exists=True), str]), multiple=True,
+              metavar='HOSTPATH PATH',
+              help="Mount a file or directory into the sandbox")
+@click.option('--isolate', is_flag=True, default=False,
+              help='Create an isolated build sandbox')
 @click.argument('element',
                 type=click.Path(dir_okay=False, readable=True))
 @click.argument('command', type=click.STRING, nargs=-1)
 @click.pass_obj
-def shell(app, element, sysroot, build, command):
+def shell(app, element, sysroot, mount, isolate, build, command):
     """Run a command in the target element's sandbox environment
 
-    This will first stage a temporary sysroot for running
-    the target element, assuming it has already been built
-    and all required artifacts are in the local cache.
+    This will stage a temporary sysroot for running the target
+    element, assuming it has already been built and all required
+    artifacts are in the local cache.
 
     Use the --build option to create a temporary sysroot for
     building the element instead.
@@ -481,9 +486,10 @@ def shell(app, element, sysroot, build, command):
     to use a specific sysroot.
 
     If no COMMAND is specified, the default is to attempt
-    to run an interactive shell with `sh -i`.
+    to run an interactive shell.
     """
     from ..element import Scope
+    from .._project import HostMount
     if build:
         scope = Scope.BUILD
     else:
@@ -507,8 +513,14 @@ def shell(app, element, sysroot, build, command):
         click.echo("Try building them first", err=True)
         sys.exit(-1)
 
+    mounts = [
+        HostMount(path, host_path)
+        for host_path, path in mount
+    ]
+
     try:
-        exitcode = app.pipeline.targets[0]._shell(scope, sysroot, command=command)
+        element = app.pipeline.targets[0]
+        exitcode = app.shell(element, scope, sysroot, mounts=mounts, isolate=isolate, command=command)
         sys.exit(exitcode)
     except BstError as e:
         click.echo("", err=True)
@@ -608,8 +620,7 @@ def workspace_open(app, no_checkout, force, track, element, directory):
         app.pipeline.open_workspace(app.scheduler, directory, no_checkout, track, force)
         click.echo("", err=True)
     except BstError as e:
-        click.echo("", err=True)
-        click.echo("ERROR: {}".format(e), err=True)
+        app.print_error(e)
         sys.exit(-1)
 
 
