@@ -22,7 +22,7 @@ import os
 from collections import OrderedDict
 from contextlib import ExitStack
 from mmap import mmap
-
+import re
 import click
 from ruamel import yaml
 
@@ -85,26 +85,45 @@ class Space(Widget):
         return ' '
 
 
+# Used to add fixed text between columns
+class FixedText(Widget):
+
+    def __init__(self, text, content_profile, format_profile):
+        super(FixedText, self).__init__(content_profile, format_profile)
+        self.text = text
+
+    def render(self, message):
+        return self.format_profile.fmt(self.text)
+
+
+# Used to add the wallclock time this message was created at
+class WallclockTime(Widget):
+    def render(self, message):
+        fields = [self.content_profile.fmt("{:02d}".format(x)) for x in
+                  [message.creation_time.hour,
+                   message.creation_time.minute,
+                   message.creation_time.second]]
+        return self.format_profile.fmt(":").join(fields)
+
+
 # A widget for rendering the debugging column
 class Debug(Widget):
 
     def render(self, message):
         unique_id = 0 if message.unique_id is None else message.unique_id
 
-        text = self.format_profile.fmt('[pid:')
+        text = self.format_profile.fmt('pid:')
         text += self.content_profile.fmt("{: <5}".format(message.pid))
         text += self.format_profile.fmt(" id:")
         text += self.content_profile.fmt("{:0>3}".format(unique_id))
-        text += self.format_profile.fmt(']')
 
         return text
 
 
 # A widget for rendering the time codes
 class TimeCode(Widget):
-
-    def __init__(self, content_profile, format_profile, brackets=True):
-        self.brackets = brackets
+    def __init__(self, content_profile, format_profile, microseconds=False):
+        self.microseconds = microseconds
         super(TimeCode, self).__init__(content_profile, format_profile)
 
     def render(self, message):
@@ -124,14 +143,13 @@ class TimeCode(Widget):
                 for field in [hours, minutes, seconds]
             ]
 
-        text = ''
-        if self.brackets:
-            text += self.format_profile.fmt('[')
-
-        text += self.format_profile.fmt(':').join(fields)
-        if self.brackets:
-            text += self.format_profile.fmt(']')
+        text = self.format_profile.fmt(':').join(fields)
 
+        if self.microseconds:
+            if elapsed is not None:
+                text += self.content_profile.fmt(".{0:06d}".format(elapsed.microseconds))
+            else:
+                text += self.content_profile.fmt(".------")
         return text
 
 
@@ -189,11 +207,9 @@ class ElementName(Widget):
         if not action_name:
             action_name = "Main"
 
-        return self.format_profile.fmt('[') + \
-            self.content_profile.fmt("{: >5}".format(action_name.lower())) + \
+        return self.content_profile.fmt("{: >5}".format(action_name.lower())) + \
             self.format_profile.fmt(':') + \
-            self.content_profile.fmt(self.fmt_string.format(name)) + \
-            self.format_profile.fmt(']')
+            self.content_profile.fmt(self.fmt_string.format(name))
 
 
 # A widget for displaying the primary message text
@@ -233,7 +249,7 @@ class CacheKey(Widget):
         else:
             text = self.content_profile.fmt(key, dim=missing)
 
-        return self.format_profile.fmt('[') + text + self.format_profile.fmt(']')
+        return text
 
 
 # A widget for formatting the log file
@@ -268,6 +284,38 @@ class LogFile(Widget):
         return text
 
 
+class SequenceID(Widget):
+    """ A unique Sequence ID is created for each timed_activity, and can
+        be used to associate all log lines from a particular
+        activity.
+    """
+    def render(self, message):
+        if message.sequence_id is None:
+            return self.content_profile.fmt("-")
+        else:
+            return self.content_profile.fmt(str(message.sequence_id))
+
+
+class MessageOrLogFile(Widget):
+    """ START and SUCCESS messages are expected to have no useful
+        information in the message text, so we display the logfile name for
+        these messages, and the message text for other types.
+    """
+    def __init__(self, content_profile, format_profile, err_profile):
+        super(MessageOrLogFile, self).__init__(content_profile, format_profile)
+        self.message_widget = MessageText(content_profile, format_profile)
+        self.logfile_widget = LogFile(content_profile, format_profile, err_profile)
+
+    def render(self, message):
+        # Show the log file only in the main start/success messages
+        if message.logfile and message.scheduler and \
+           message.message_type in [MessageType.START, MessageType.SUCCESS]:
+            text = self.logfile_widget.render(message)
+        else:
+            text = self.message_widget.render(message)
+        return text
+
+
 # A widget for formatting a log line
 class LogLine(Widget):
 
@@ -275,7 +323,8 @@ class LogLine(Widget):
                  indent=4,
                  log_lines=10,
                  message_lines=10,
-                 debug=False):
+                 debug=False,
+                 message_format: str=None):
         super(LogLine, self).__init__(content_profile, format_profile)
 
         self.columns = []
@@ -285,9 +334,9 @@ class LogLine(Widget):
         self.indent = ' ' * indent
         self.log_lines = log_lines
         self.message_lines = message_lines
+        self.message_format = message_format
 
         self.space_widget = Space(content_profile, format_profile)
-        self.message_widget = MessageText(content_profile, format_profile)
         self.logfile_widget = LogFile(content_profile, format_profile, err_profile)
 
         if debug:
@@ -295,21 +344,51 @@ class LogLine(Widget):
                 Debug(content_profile, format_profile)
             ])
 
-        self.columns.extend([
-            TimeCode(content_profile, format_profile),
-            CacheKey(content_profile, format_profile, err_profile),
-            ElementName(content_profile, format_profile),
-            self.space_widget,
-            TypeName(content_profile, format_profile),
-            self.space_widget
-        ])
+        logfile_format = message_format
+
+        self.logfile_variable_names = {
+            "elapsed": TimeCode(content_profile, format_profile, microseconds=False),
+            "elapsed-us": TimeCode(content_profile, format_profile, microseconds=True),
+            "wallclock": WallclockTime(content_profile, format_profile),
+            "key": CacheKey(content_profile, format_profile, err_profile),
+            "element": ElementName(content_profile, format_profile),
+            "action": TypeName(content_profile, format_profile),
+            "message": MessageOrLogFile(content_profile, format_profile, err_profile),
+            "sequence": SequenceID(content_profile, format_profile)
+        }
+        logfile_tokens = self._parse_logfile_format(logfile_format, content_profile, format_profile)
+        self.columns.extend(logfile_tokens)
+
+    def _parse_logfile_format(self, format_string, content_profile, format_profile):
+        logfile_tokens = []
+        while len(format_string) > 0:
+            if format_string.startswith("%%"):
+                logfile_tokens.append(FixedText("%", content_profile, format_profile))
+                format_string = format_string[2:]
+                continue
+            m = re.search("^%\{([^\}]+)\}", format_string)
+            if m is not None:
+                variable = m.group(1)
+                format_string = format_string[m.end(0):]
+                if variable not in self.logfile_variable_names:
+                    raise Exception("'{0}' is not a valid log variable name.".format(variable))
+                logfile_tokens.append(self.logfile_variable_names[variable])
+            else:
+                m = re.search("^[^%]+", format_string)
+                if m is not None:
+                    text = FixedText(m.group(0), content_profile, format_profile)
+                    format_string = format_string[m.end(0):]
+                    logfile_tokens.append(text)
+                else:
+                    # No idea what to do now
+                    raise Exception("'{0}' could not be parsed into a valid logging format.".format(format_string))
+        return logfile_tokens
 
     def size_request(self, pipeline):
         for widget in self.columns:
             widget.size_request(pipeline)
 
         self.space_widget.size_request(pipeline)
-        self.message_widget.size_request(pipeline)
         self.logfile_widget.size_request(pipeline)
 
     def render(self, message):
@@ -319,13 +398,6 @@ class LogLine(Widget):
         for widget in self.columns:
             text += widget.render(message)
 
-        # Show the log file only in the main start/success messages
-        if message.logfile and message.scheduler and \
-           message.message_type in [MessageType.START, MessageType.SUCCESS]:
-            text += self.logfile_widget.render(message)
-        else:
-            text += self.message_widget.render(message)
-
         text += '\n'
 
         extra_nl = False
