@@ -164,7 +164,7 @@ class Pipeline():
     def preflight(self):
         for plugin in self.dependencies(Scope.ALL, include_sources=True):
             try:
-                plugin.preflight()
+                plugin._preflight()
             except BstError as e:
                 # Prepend the plugin identifier string to the error raised by
                 # the plugin so that users can more quickly identify the issue
@@ -561,16 +561,12 @@ class Pipeline():
         workdir = os.path.abspath(directory)
 
         if len(list(target.sources())) == 0:
-            raise PipelineError("The given element has no sources")
-
-        # Check directory
-        try:
-            os.makedirs(directory, exist_ok=True)
-        except OSError as e:
-            raise PipelineError("Failed to create workspace directory: {}".format(e)) from e
-
-        if not no_checkout and not force and os.listdir(directory):
-            raise PipelineError("Checkout directory is not empty: {}".format(directory))
+            build_depends = [x.name for x in target.dependencies(Scope.BUILD, recurse=False)]
+            if len(build_depends) == 0:
+                raise PipelineError("The given element has no sources or build-dependencies")
+            detail = "Try opening a workspace on one of its dependencies instead:\n"
+            detail += "  \n".join(build_depends)
+            raise PipelineError("The given element has no sources", detail=detail)
 
         # Check for workspace config
         if self.project._get_workspace(target.name):
@@ -608,14 +604,24 @@ class Pipeline():
                 self.message(MessageType.SUCCESS,
                              "Fetched {} elements".format(fetched), elapsed=elapsed)
 
+        if not no_checkout and target._consistency() != Consistency.CACHED:
+            raise PipelineError("Could not stage uncached source. " +
+                                "Use `--track` to track and " +
+                                "fetch the latest version of the " +
+                                "source.")
+
+        # Check directory
+        try:
+            os.makedirs(directory, exist_ok=True)
+        except OSError as e:
+            raise PipelineError("Failed to create workspace directory: {}".format(e)) from e
+
         if not no_checkout:
+            if not force and os.listdir(directory):
+                raise PipelineError("Checkout directory is not empty: {}".format(directory))
+
             with target.timed_activity("Staging sources to {}".format(directory)):
                 for source in target.sources():
-                    if source.get_consistency() != Consistency.CACHED:
-                        raise PipelineError("Could not stage uncached source. " +
-                                            "Use `--track` to track and " +
-                                            "fetch the latest version of the " +
-                                            "source.")
                     source._init_workspace(directory)
 
         self.project._set_workspace(target, workdir)
