@@ -37,14 +37,36 @@ from ._sourcefactory import SourceFactory
 # The base BuildStream format version
 #
 # This version is bumped whenever enhancements are made
-# to the ``project.conf`` format or the format in general.
+# to the `project.conf` format or the core element format.
 #
-BST_FORMAT_VERSION = 0
+BST_FORMAT_VERSION = 4
 
 # The separator we use for user specified aliases
 _ALIAS_SEPARATOR = ':'
 
 
+# HostMount()
+#
+# A simple object describing the behavior of
+# a host mount.
+#
+class HostMount():
+
+    def __init__(self, path, host_path=None, optional=False):
+
+        # Support environment variable expansion in host mounts
+        path = os.path.expandvars(path)
+        if host_path is not None:
+            host_path = os.path.expandvars(host_path)
+
+        self.path = path              # Path inside the sandbox
+        self.host_path = host_path    # Path on the host
+        self.optional = optional      # Optional mounts do not incur warnings or errors
+
+        if self.host_path is None:
+            self.host_path = self.path
+
+
 # Project()
 #
 # The Project Configuration
@@ -66,6 +88,7 @@ class Project():
         self._variables = {}     # The default variables overridden with project wide overrides
         self._environment = {}   # The base sandbox environment
         self._elements = {}      # Element specific configurations
+        self._sources = {}       # Source specific configurations
         self._aliases = {}       # Aliases dictionary
         self._workspaces = {}    # Workspaces
         self._plugin_source_origins = []   # Origins of custom sources
@@ -78,6 +101,11 @@ class Project():
         self._element_format_versions = {}
         self._fail_on_overlap = False
 
+        # Shell options
+        self._shell_command = []      # The default interactive shell command
+        self._shell_environment = {}  # Statically set environment vars
+        self._shell_host_files = []   # A list of HostMount objects
+
         profile_start(Topics.LOAD_PROJECT, self.directory.replace(os.sep, '-'))
         self._load()
         profile_end(Topics.LOAD_PROJECT, self.directory.replace(os.sep, '-'))
@@ -123,10 +151,13 @@ class Project():
         project_conf = _yaml.load(projectfile)
         _yaml.composite(config, project_conf)
 
-        # Element type configurations will be composited later onto element types,
-        # so we delete it from here and run our final assertion after.
+        # Element and Source  type configurations will be composited later onto
+        # element/source types, so we delete it from here and run our final
+        # assertion after.
         self._elements = _yaml.node_get(config, Mapping, 'elements', default_value={})
+        self._sources = _yaml.node_get(config, Mapping, 'sources', default_value={})
         config.pop('elements', None)
+        config.pop('sources', None)
         _yaml.node_final_assertions(config)
         _yaml.node_validate(config, [
             'format-version',
@@ -135,7 +166,7 @@ class Project():
             'split-rules', 'elements', 'plugins',
             'aliases', 'name',
             'artifacts', 'options',
-            'fail-on-overlap'
+            'fail-on-overlap', 'shell'
         ])
 
         # The project name, element path and option declarations
@@ -265,6 +296,37 @@ class Project():
         self._fail_on_overlap = _yaml.node_get(config, bool, 'fail-on-overlap',
                                                default_value=False)
 
+        # Parse shell options
+        shell_options = _yaml.node_get(config, Mapping, 'shell', default_value={})
+        _yaml.node_validate(shell_options, ['command', 'environment', 'host-files'])
+        self._shell_command = _yaml.node_get(shell_options, list, 'command',
+                                             default_value=['sh', '-i'])
+
+        # Perform environment expansion right away
+        shell_environment = _yaml.node_get(shell_options, Mapping, 'environment', default_value={})
+        for key, _ in _yaml.node_items(shell_environment):
+            value = _yaml.node_get(shell_environment, str, key)
+            self._shell_environment[key] = os.path.expandvars(value)
+
+        # Host files is parsed as a list for convenience
+        host_files = _yaml.node_get(shell_options, list, 'host-files', default_value=[])
+        for host_file in host_files:
+            if isinstance(host_file, str):
+                mount = HostMount(host_file)
+            else:
+                # Some validation
+                index = host_files.index(host_file)
+                host_file_desc = _yaml.node_get(shell_options, Mapping, 'host-files', indices=[index])
+                _yaml.node_validate(host_file_desc, ['path', 'host_path', 'optional'])
+
+                # Parse the host mount
+                path = _yaml.node_get(host_file_desc, str, 'path')
+                host_path = _yaml.node_get(host_file_desc, str, 'host_path', default_value='') or None
+                optional = _yaml.node_get(host_file_desc, bool, 'optional', default_value=False)
+                mount = HostMount(path, host_path, optional)
+
+            self._shell_host_files.append(mount)
+
     # _store_origin()
     #
     # Helper function to store plugin origins
