@@ -19,9 +19,6 @@
 #        Tristan Van Berkom <tristan.vanberkom@codethink.co.uk>
 #        Jürg Billeter <juerg.billeter@codethink.co.uk>
 
-# BuildStream toplevel imports
-from ..element import _KeyStrength
-
 # Local imports
 from . import Queue, QueueStatus, QueueType
 
@@ -42,8 +39,7 @@ class PullQueue(Queue):
         # state of dependencies may have changed, recalculate element state
         element._update_state()
 
-        # cache cannot be queried until strict cache key is available
-        if element._get_strict_cache_key() is None:
+        if not element._can_query_cache():
             return QueueStatus.WAIT
 
         if element._pull_pending():
