@@ -25,8 +25,12 @@ from contextlib import ExitStack
 
 from ruamel import yaml
 from ruamel.yaml.representer import SafeRepresenter, RoundTripRepresenter
+from ruamel.yaml.constructor import RoundTripConstructor
 from ._exceptions import LoadError, LoadErrorReason
 
+# This overrides the ruamel constructor to treat everything as a string
+RoundTripConstructor.add_constructor(u'tag:yaml.org,2002:int', RoundTripConstructor.construct_yaml_str)
+RoundTripConstructor.add_constructor(u'tag:yaml.org,2002:float', RoundTripConstructor.construct_yaml_str)
 
 # We store information in the loaded yaml on a DictProvenance
 # stored in all dictionaries under this key
@@ -187,7 +191,7 @@ def load(filename, shortname=None, copy_tree=False):
 def load_data(data, shortname=None, copy_tree=False):
 
     try:
-        contents = yaml.load(data, yaml.loader.RoundTripLoader)
+        contents = yaml.load(data, yaml.loader.RoundTripLoader, preserve_quotes=True)
     except (yaml.scanner.ScannerError, yaml.composer.ComposerError, yaml.parser.ParserError) as e:
         raise LoadError(LoadErrorReason.INVALID_YAML,
                         "Malformed YAML:\n\n{}\n\n{}\n".format(e.problem, e.problem_mark)) from e
@@ -213,7 +217,8 @@ def load_data(data, shortname=None, copy_tree=False):
 def dump(node, filename=None):
     with ExitStack() as stack:
         if filename:
-            f = stack.enter_context(open(filename, 'w'))
+            from . import utils
+            f = stack.enter_context(utils.save_file_atomic(filename, 'w'))
         else:
             f = sys.stdout
         yaml.round_trip_dump(node, f)
