@@ -128,6 +128,14 @@ class Element(Plugin):
     any of the dependencies have changed.
     """
 
+    BST_FORBID_RDEPENDS = False
+    """Whether to raise exceptions if an element has runtime-dependencies.
+    """
+
+    BST_FORBID_SOURCES = False
+    """Whether to raise exceptions if an element has sources.
+    """
+
     def __init__(self, context, project, artifacts, meta, plugin_conf):
 
         super().__init__(meta.name, context, project, meta.provenance, "element")
@@ -152,6 +160,8 @@ class Element(Plugin):
         self.__strong_cached = None             # Whether we have a cached artifact
         self.__remotely_cached = None           # Whether we have a remotely cached artifact
         self.__remotely_strong_cached = None    # Whether we have a remotely cached artifact
+        self.__assemble_scheduled = False       # Element is scheduled to be assembled
+        self.__assemble_done = False            # Element is assembled
         self.__pull_failed = False              # Whether pull was attempted but failed
         self.__log_path = None                  # Path to dedicated log file or None
         self.__splits = None
@@ -421,7 +431,7 @@ class Element(Plugin):
 
         with self.timed_activity("Staging {}/{}".format(self.name, self._get_display_key())):
             # Get the extracted artifact
-            artifact = os.path.join(self.__artifacts.extract(self), 'files')
+            artifact = os.path.join(self.__extract(), 'files')
 
             # Hard link it into the staging area
             #
@@ -787,6 +797,9 @@ class Element(Plugin):
     # in a subprocess.
     #
     def _schedule_assemble(self):
+        assert(not self.__assemble_scheduled)
+        self.__assemble_scheduled = True
+
         for source in self.__sources:
             source._schedule_assemble()
 
@@ -798,9 +811,14 @@ class Element(Plugin):
     # in a subprocess.
     #
     def _assemble_done(self):
+        assert(self.__assemble_scheduled)
+
         for source in self.__sources:
             source._assemble_done()
 
+        self.__assemble_scheduled = False
+        self.__assemble_done = True
+
         self._update_state()
 
     # _cached():
@@ -809,17 +827,8 @@ class Element(Plugin):
     #    (bool): Whether this element is already present in
     #            the artifact cache
     #
-    def _cached(self, strength=None):
-
-        if strength is None:
-            strength = _KeyStrength.STRONG if self._get_strict() else _KeyStrength.WEAK
-
-        if self.__cached is None:
-            return False
-        elif strength == _KeyStrength.STRONG:
-            return self.__strong_cached
-        else:
-            return self.__cached
+    def _cached(self):
+        return self.__cached
 
     # _assert_cached()
     #
@@ -836,17 +845,8 @@ class Element(Plugin):
     #    (bool): Whether this element is already present in
     #            the remote artifact cache
     #
-    def _remotely_cached(self, strength=None):
-
-        if strength is None:
-            strength = _KeyStrength.STRONG if self._get_strict() else _KeyStrength.WEAK
-
-        if self.__remotely_cached is None:
-            return False
-        elif strength == _KeyStrength.STRONG:
-            return self.__remotely_strong_cached
-        else:
-            return self.__remotely_cached
+    def _remotely_cached(self):
+        return self.__remotely_cached
 
     # _tainted():
     #
@@ -964,17 +964,24 @@ class Element(Plugin):
         else:
             return self.__weak_cache_key
 
-    # _get_strict_cache_key():
+    # _can_query_cache():
     #
-    # Returns the cache key for strict build mode
+    # Returns whether the cache key required for cache queries is available.
     #
     # Returns:
-    #    (str): A hex digest cache key for this Element, or None
+    #    (bool): True if cache can be queried
     #
-    # None is returned if information for the cache key is missing.
-    #
-    def _get_strict_cache_key(self):
-        return self.__strict_cache_key
+    def _can_query_cache(self):
+        # If build has already been scheduled, we know that the element is
+        # not cached and thus can allow cache query even if the strict cache key
+        # is not available yet.
+        # This special case is required for workspaced elements to prevent
+        # them from getting blocked in the pull queue.
+        if self.__assemble_scheduled:
+            return True
+
+        # cache cannot be queried until strict cache key is available
+        return self.__strict_cache_key is not None
 
     # _get_full_display_key():
     #
@@ -1136,7 +1143,7 @@ class Element(Plugin):
                 _yaml.dump(_yaml.node_sanitize(meta), os.path.join(metadir, 'artifact.yaml'))
 
                 with self.timed_activity("Caching Artifact"):
-                    self.__artifacts.commit(self, assembledir)
+                    self.__artifacts.commit(self, assembledir, self.__get_cache_keys_for_commit())
 
             # Finally cleanup the build dir
             cleanup_rootdir()
@@ -1152,7 +1159,26 @@ class Element(Plugin):
         def progress(percent, message):
             self.status(message)
 
-        self.__artifacts.pull(self, progress=progress)
+        weak_key = self._get_cache_key(strength=_KeyStrength.WEAK)
+
+        if self.__remotely_strong_cached:
+            key = self.__strict_cache_key
+            self.__artifacts.pull(self, key, progress=progress)
+
+            # update weak ref by pointing it to this newly fetched artifact
+            self.__artifacts.link_key(self, key, weak_key)
+        elif not self._get_strict() and self.__remotely_cached:
+            self.__artifacts.pull(self, weak_key, progress=progress)
+
+            # extract strong cache key from this newly fetched artifact
+            self._update_state()
+
+            # create tag for strong cache key
+            key = self._get_cache_key(strength=_KeyStrength.STRONG)
+            self.__artifacts.link_key(self, weak_key, key)
+        else:
+            raise ElementError("Attempt to pull unavailable artifact for element {}"
+                               .format(self.name))
 
         # Notify successfull download
         display_key = self._get_display_key()
@@ -1161,10 +1187,10 @@ class Element(Plugin):
 
     # _skip_push():
     #
-    # Determine whether element should be pushed.
+    # Determine whether we should create a push job for this element.
     #
     # Returns:
-    #   (bool): True if this element should not be pushed
+    #   (bool): True if this element does not need a push job to be created
     #
     def _skip_push(self):
         if not self.__artifacts.has_push_remotes(element=self):
@@ -1178,7 +1204,16 @@ class Element(Plugin):
         if self._tainted():
             return True
 
-        return False
+        # Use the strong cache key to check whether a remote already has the artifact.
+        # In non-strict mode we want to push updated artifacts even if the
+        # remote already has an artifact with the same weak cache key.
+        key = self._get_cache_key(strength=_KeyStrength.STRONG)
+
+        # Skip if every push remote contains this element already.
+        if self.__artifacts.push_needed(self, key):
+            return False
+        else:
+            return True
 
     # _push():
     #
@@ -1196,7 +1231,8 @@ class Element(Plugin):
             return False
 
         with self.timed_activity("Pushing Artifact"):
-            return self.__artifacts.push(self)
+            # Push all keys used for local commit
+            return self.__artifacts.push(self, self.__get_cache_keys_for_commit())
 
     # _logfile()
     #
@@ -1265,7 +1301,7 @@ class Element(Plugin):
         if self.__workspaced_artifact is None:
             self._assert_cached()
 
-            metadir = os.path.join(self.__artifacts.extract(self), 'meta')
+            metadir = os.path.join(self.__extract(), 'meta')
             meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
             if 'workspaced' in meta:
                 self.__workspaced_artifact = meta['workspaced']
@@ -1279,7 +1315,7 @@ class Element(Plugin):
         if self.__workspaced_dependencies_artifact is None:
             self._assert_cached()
 
-            metadir = os.path.join(self.__artifacts.extract(self), 'meta')
+            metadir = os.path.join(self.__extract(), 'meta')
             meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
             if 'workspaced_dependencies' in meta:
                 self.__workspaced_dependencies_artifact = meta['workspaced_dependencies']
@@ -1368,39 +1404,56 @@ class Element(Plugin):
     # Args:
     #    scope (Scope): Either BUILD or RUN scopes are valid, or None
     #    directory (str): A directory to an existing sandbox, or None
+    #    mounts (list): A list of (str, str) tuples, representing host/target paths to mount
+    #    isolate (bool): Whether to isolate the environment like we do in builds
+    #    prompt (str): A suitable prompt string for PS1
     #    command (list): An argv to launch in the sandbox
     #
     # Returns: Exit code
     #
     # If directory is not specified, one will be staged using scope
-    def _shell(self, scope=None, directory=None, command=None):
+    def _shell(self, scope=None, directory=None, *, mounts=None, isolate=False, prompt=None, command=None):
 
         with self._prepare_sandbox(scope, directory) as sandbox:
             environment = self.get_environment()
-            flags = SandboxFlags.INTERACTIVE
+            environment = copy.copy(environment)
+            flags = SandboxFlags.INTERACTIVE | SandboxFlags.ROOT_READ_ONLY
 
-            if scope == Scope.RUN:
-                # If a testing sandbox was requested, override the element environment
-                # with some of the host environment and use that for the shell.
+            # Fetch the main toplevel project, in case this is a junctioned
+            # subproject, we want to use the rules defined by the main one.
+            context = self._get_context()
+            project = context._get_toplevel_project()
+
+            if prompt is not None:
+                environment['PS1'] = prompt
+
+            # Special configurations for non-isolated sandboxes
+            if not isolate:
+
+                # Open the network, and reuse calling uid/gid
                 #
-                # XXX Hard code should be removed
-                environment = copy.copy(environment)
-                overrides = ['DISPLAY', 'DBUS_SESSION_BUS_ADDRESS']
-                for override in overrides:
-                    if os.environ.get(override) is not None:
-                        environment[override] = os.environ.get(override)
+                flags |= SandboxFlags.NETWORK_ENABLED | SandboxFlags.INHERIT_UID
 
-            # Decide whether to create a build style sandbox or an open
-            # testing sandbox based on whether the build scope was requested
-            if scope == Scope.BUILD:
-                flags |= SandboxFlags.ROOT_READ_ONLY
-            elif scope == Scope.RUN:
-                flags |= SandboxFlags.NETWORK_ENABLED
+                # Apply project defined environment vars to set for a shell
+                for key, value in _yaml.node_items(project._shell_environment):
+                    environment[key] = value
+
+                # Setup any requested bind mounts
+                if mounts is None:
+                    mounts = []
+
+                for mount in project._shell_host_files + mounts:
+                    if not os.path.exists(mount.host_path):
+                        if not mount.optional:
+                            self.warn("Not mounting non-existing host file: {}".format(mount.host_path))
+                    else:
+                        sandbox.mark_directory(mount.path)
+                        sandbox._set_mount_source(mount.path, mount.host_path)
 
             if command:
                 argv = [arg for arg in command]
             else:
-                argv = ['sh', '-i']
+                argv = project._shell_command
 
             self.status("Running command", detail=" ".join(argv))
 
@@ -1552,6 +1605,20 @@ class Element(Plugin):
                 # Weak cache key could not be calculated yet
                 return
 
+        if not self._get_strict():
+            # Full cache query in non-strict mode requires both the weak and
+            # strict cache keys. However, we need to determine as early as
+            # possible whether a build is pending to discard unstable cache keys
+            # for workspaced elements. For this cache check the weak cache keys
+            # are sufficient. However, don't update the `cached` attributes
+            # until the full cache query below.
+            cached = self.__artifacts.contains(self, self.__weak_cache_key)
+            remotely_cached = self.__artifacts.remote_contains(self, self.__weak_cache_key)
+            if (not self.__assemble_scheduled and not self.__assemble_done and
+                not cached and not remotely_cached):
+                self._schedule_assemble()
+                return
+
         if self.__strict_cache_key is None:
             dependencies = [
                 e.__strict_cache_key for e in self.dependencies(Scope.BUILD)
@@ -1563,14 +1630,24 @@ class Element(Plugin):
                 return
 
         # Query caches now that the weak and strict cache keys are available
+        key_for_cache_lookup = self.__strict_cache_key if self._get_strict() else self.__weak_cache_key
         if not self.__cached:
-            self.__cached = self.__artifacts.contains(self)
+            self.__cached = self.__artifacts.contains(self, key_for_cache_lookup)
         if not self.__remotely_cached:
-            self.__remotely_cached = self.__artifacts.remote_contains(self)
+            self.__remotely_cached = self.__artifacts.remote_contains(self, key_for_cache_lookup)
         if not self.__strong_cached:
-            self.__strong_cached = self.__artifacts.contains(self, strength=_KeyStrength.STRONG)
+            self.__strong_cached = self.__artifacts.contains(self, self.__strict_cache_key)
         if not self.__remotely_strong_cached:
-            self.__remotely_strong_cached = self.__artifacts.remote_contains(self, strength=_KeyStrength.STRONG)
+            self.__remotely_strong_cached = self.__artifacts.remote_contains(self, self.__strict_cache_key)
+
+        if (not self.__assemble_scheduled and not self.__assemble_done and
+            not self.__cached and not self.__remotely_cached):
+            # Workspaced sources are considered unstable if a build is pending
+            # as the build will modify the contents of the workspace.
+            # Determine as early as possible if a build is pending to discard
+            # unstable cache keys.
+            self._schedule_assemble()
+            return
 
         if self.__cache_key is None:
             # Calculate strong cache key
@@ -1581,7 +1658,7 @@ class Element(Plugin):
                 pass
             elif self._cached():
                 # Load the strong cache key from the artifact
-                metadir = os.path.join(self.__artifacts.extract(self), 'meta')
+                metadir = os.path.join(self.__extract(), 'meta')
                 meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
                 self.__cache_key = meta['keys']['strong']
             elif self._buildable():
@@ -1595,6 +1672,21 @@ class Element(Plugin):
                 # Strong cache key could not be calculated yet
                 return
 
+    def _preflight(self):
+        if self.BST_FORBID_RDEPENDS:
+            runtime_deps = list(self.dependencies(Scope.RUN, recurse=False))
+            if runtime_deps:
+                raise ElementError("{}: Runtime dependencies are forbidden for elements of type {}"
+                                   .format(self, type(self).__name__), reason="element-forbidden-rdepends")
+
+        if self.BST_FORBID_SOURCES:
+            sources = list(self.sources())
+            if sources:
+                raise ElementError("{}: Sources are forbidden for elements of type {}"
+                                   .format(self, type(self).__name__), reason="element-forbidden-sources")
+
+        self.preflight()
+
     #############################################################
     #                   Private Local Methods                   #
     #############################################################
@@ -1781,7 +1873,7 @@ class Element(Plugin):
         }
 
     def __compute_splits(self, include=None, exclude=None, orphans=True):
-        basedir = os.path.join(self.__artifacts.extract(self), 'files')
+        basedir = os.path.join(self.__extract(), 'files')
 
         # No splitting requested, just report complete artifact
         if orphans and not (include or exclude):
@@ -1831,12 +1923,33 @@ class Element(Plugin):
             if include_file and not exclude_file:
                 yield filename.lstrip(os.sep)
 
+    def __extract(self):
+        key = self.__strict_cache_key
+
+        # Use weak cache key, if artifact is missing for strong cache key
+        # and the context allows use of weak cache keys
+        if not self._get_strict() and not self.__artifacts.contains(self, key):
+            key = self._get_cache_key(strength=_KeyStrength.WEAK)
+
+        return self.__artifacts.extract(self, key)
+
+    def __get_cache_keys_for_commit(self):
+        keys = []
+
+        # tag with strong cache key based on dependency versions used for the build
+        keys.append(self._get_cache_key(strength=_KeyStrength.STRONG))
+
+        # also store under weak cache key
+        keys.append(self._get_cache_key(strength=_KeyStrength.WEAK))
+
+        return utils._deduplicate(keys)
+
     def _load_public_data(self):
         self._assert_cached()
         assert(self.__dynamic_public is None)
 
         # Load the public data from the artifact
-        metadir = os.path.join(self.__artifacts.extract(self), 'meta')
+        metadir = os.path.join(self.__extract(), 'meta')
         self.__dynamic_public = _yaml.load(os.path.join(metadir, 'public.yaml'))
 
     def _subst_string(self, value):
