@@ -1,4 +1,5 @@
-from buildstream._options import OptionPool
+from buildstream._context import Context
+from buildstream._project import Project
 
 
 #
@@ -6,7 +7,5 @@ from buildstream._options import OptionPool
 # be removed in favor of testing the functionality via
 # the CLI like in the frontend tests anyway.
 #
-def make_options(basedir):
-    options = OptionPool(basedir)
-    options.resolve()
-    return options
+def make_project(basedir):
+    return Project(basedir, Context())
